/* File: comm-unix.cc
 *
 *     This file is part of minkowsky
 *
 *     Copyright (C) 2001-2002 by Rdiger Goetz
 *     Author: Rdiger Goetz <minkowsky@r-goetz.de>
 *
 *     Time-stamp: <27-Oct-2002 15:52:45 goetz>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "../client.h"


// Functions in this module used outside the module:
//
// void   initSockets()
// char * logIntoServer(char *curUserName, char *curPasswd, int reconnect )
// int    openSocket(char *master,int port)
// int    send2Socket(int socket,char *msg)
// int    send2DamagedSocket(int socket,char *msg)
// int    wait4Message(int socket, char *buffer,int len, int timeout)
// int    send2Server(char *cmd, char *buffer, int bufLen)
// int    getAnswer(char **buffer, char * ticket)
// char * checkServer()
// int    closeSocket ()
// void closeApp(int value)


void cleanMessagePipe(int socket);
int extendBuffer(char **t, int bufLen,int newLen);
char *wait4Answer(int socket, int *succ);
void pipebroken (int value);


void initSockets()
{
#ifdef NO_POLL
 signal(SIGALRM,timerStop);  
#ifdef MACOSX
 signal(SIGIO, setGotData);
#else
 sigioAction.sa_handler   = NULL;
 sigioAction.sa_restorer  = NULL;
 sigioAction.sa_sigaction = setGotData;
 sigemptyset(&sigioAction.sa_mask);
 sigioAction.sa_flags     = SA_SIGINFO;
 sigaction(SIGIO,&sigioAction,NULL);
#endif
#else
 signal(SIGIO, SIG_IGN);
#endif
 signal(SIGPIPE, pipebroken);
}


#ifdef WITH_CLC // FIXME fix broken command line client

//
// Function to log into the Server fpr the command line version of the client
//
int cl_logIntoServer()
{
 char *cp[4],cdata[4][256],res[1024];
//  ClientData cData;
//  Tcl_Interp interp;

 printf("Minkowsky Commandline client\n\n");

 cp[0]=cdata[0];
 cp[1]=cdata[1];
 cp[2]=cdata[2];

 if(username[0] == '\0' || password[0]=='\0')
  printf("you need to authenticate to login\n");
 
 if(username[0] == '\0')
  {
   printf("Username: ");
   strcpy(cdata[0],"login");
   fgets(cdata[1],255,stdin);
   printf("\n");
  }
 else
  cp[1] = username;

 if(username[0] == '\0' || password[0]=='\0')
  strcpy(cdata[2],getpass("Password: "));
 else
  cp[2] =    password ;

 interp.result=res;
 logIntoServer(cData,&interp,2,cp);
 return(0);
}
#endif


//
// Function to log into the server
// Error (if any) is return in inter->result
//
char * logIntoServer(char *curUserName, char *curPasswd, int reconnect )
{
 char cmd[256],*buffer;
 char buf[1024];
 static char *txt=NULL;
 static int lc=0;
 int retries=4;
 int myCid;

 if(txt == NULL)
  txt = new char [4096];


 do
  {
   mySocket= openSocket(terminHost,MASTERPORT);
   if(mySocket <0)
    {
     if(verbose)
      printf("Can't get master socket\n");
     retries --;
     usleep(250000);
    }
  }while(mySocket<0 && retries >0);
 if(mySocket<0)
  return "Error: {Server not responding}";

 if(reconnect)
  sprintf(cmd,"reconnect %s %s %d",curUserName,curPasswd,lc);
 else
  sprintf(cmd,"connect %s %s %d",curUserName,curPasswd,lc);
 lc++;
 send2Socket(mySocket,cmd);
 retries=3;
 int slp=500000;
 do
  {
   if ( wait4Message(mySocket,buf,1023,1000) )
    break;

   if(verbose)
    printf("retry %d %d\n",retries,slp);
   retries --;
   usleep(slp);
   slp *=2;
  }while(retries >0);
 buffer=buf;
 if(verbose)
  printf(">Got as connect reply: '%s' from %s:%d\n",buffer,terminHost,MASTERPORT);

 if(buffer[0] == '\0')
  return "Error: {Server not responding}";
 if ( strcmp(buffer,"denied")==0)
  return "Error: {Keine Zugangsberechtigung}";
 if ( strcmp(buffer,"noResource")==0)
  return "Error {Server berlastet}";
 if ( strcmp(buffer,"noConnection")==0)
  return "Error (Port berlastet)}";

 if ( strncmp(buffer,"Error:",6)==0)
  return buffer+7; 
 if( strncmp(buffer,"connected",9)!= 0 )
  {
   if(verbose)
    printf("Error : Connection to server '%s' broken\nServer doesn't answer correctly\n(answer='%s' instead of 'connected')",terminHost,buffer);
   sprintf(txt,"Error : Connection to server '%s' broken\nServer doesn't answer correctly\n(answer='%s' instead of 'connected cid')",terminHost,buffer);
   return txt;
  }
 send2Socket(mySocket,"client connected");
 myCid=atoi(buffer+9);
 if(verbose)
  printf("sent 'client connected on cid=%d'\n",myCid);
 wait4Message(mySocket,buf,1023,2000);

 if(verbose)
  printf(">> Reply: '%s'\n",buffer);
 // socketsUp=true;
 return "OK";
}


//
// Open the socket to the server 'master' on port 'port'
//
int openSocket(char *master,int port)
{
 struct hostent     *he;
 struct sockaddr_in  server;
 int ret,flg;
 int lsocket=-1;

 lsocket=socket(AF_INET,SOCK_STREAM,0);

 if ((he = gethostbyname(master)) == NULL)
  {
   puts("error resolving hostname..");
   exit(1);
  }

 memcpy(&server.sin_addr, he->h_addr_list[0], he->h_length);
 server.sin_family = AF_INET;
 server.sin_port = htons(port);

 if( (ret=connect(lsocket, (struct sockaddr *)&server, sizeof(server))) != 0)
  {
   if(verbose)
    printf("Can't establish connection to server \n");
   return(-1);
  }
 
 flg = fcntl(lsocket, F_SETOWN,getpid());

#ifdef NO_POLL
#ifndef MACOSX
 flg = fcntl(lsocket, F_SETSIG,SIGIO);
#endif
#endif

 flg = fcntl(lsocket, F_SETFL, O_NONBLOCK);
 if(verbose)
  printf("ret=%d, lsocket=%d\n",ret,lsocket);
 return(lsocket);
}


//
// Send a message to server via the socket
//
int send2Socket(int socket,char *msg)
{
 int ret;
 if(verbose)
  printf("sending on %d '%s'\n",socket,msg);
 ret=send(socket,msg,strlen(msg)+1,0);
 return(ret);
}


//
// Send a message to a socket which connection is presuable down, hopeing that
// the server may receive it.
//
int send2DamagedSocket(int socket,char *msg)
{
 int ret;
 if(verbose)
  printf("sending on %d '%s' (NOSIG)\n",socket,msg);
#ifdef MACOSX
 ret=send(socket,msg,strlen(msg)+1,0);
#else  
 ret=send(socket,msg,strlen(msg)+1,MSG_NOSIGNAL);
#endif
 return(ret);
}

//
// Clean all data for the open socket in order to get a clean response on a request
//
void cleanMessagePipe(int socket)
{
#ifndef MACOSX
 char **inp,t[65536];
 int ret;
 struct pollfd ps={socket,POLLIN,0};
 int timeout=1;

 if((ret=poll(&ps,(unsigned int) 1,timeout))>=0)
  {
   ret =recv(socket,t,65535,0);
  }
#endif
}

//
// Extend the char-buffer on *t from bufLen to newLen
// 
//
int extendBuffer(char **t, int bufLen,int newLen)
{
 char *buf;

 buf = new char [newLen];
 if(buf == NULL)
  return bufLen;

 memcpy(buf,*t,bufLen*sizeof(char));
 delete *t;
 *t = buf;
 return(newLen);
}


//
// Wait for an answer on a request (usually a 'getAnswer' request)
// called by getAnser
//
char *wait4Answer(int socket, int *succ)
{
 static char *buffer=NULL;
 static int   bufLen=0;
 char *buf,*t,*s;
 char block[256];
 int size,ret,tl=0,ttl=10,st=2000;
 struct pollfd ps={socket,POLLIN,0};

 if(buffer == NULL)
  {
   buffer = new char [4096];
   bufLen = 4096;
  }

 if(verbose)
  printf("poll fo data (timeout=%d)\n",500);
 if((ret=poll(&ps,(unsigned int) 1,1000))<1)
  {
   *succ =false;
   if(verbose)
    printf("poll failed %d\n",ret);
   return "no data";
  }

 ret =recv(socket,block,255,0);
 if( (t= strstr(block,"pending")) != NULL)
  {
   *succ=false;
   return "pending";
  }

 
 t= strstr(block,"%#");
 if(t==NULL)
  {
   *succ=false;
   return "";
  }
 s = strchr(t+2,'#');
 if(s==NULL)
  {
   *succ=false;
   return "";
  }

 *s='\0';
 s++;
 size = atoi(t+2);
 if(size+256  > bufLen)
  bufLen = extendBuffer(&buffer,bufLen,size+256);
 if(verbose)
  printf("got size %d, \t\tstarting poll2 \n",size);
 buffer[0]='\0';

 if((ret=poll(&ps,(unsigned int) 1,100))<1)
  {
   *succ =false;
   if(verbose)
    printf("poll2 failed %d\n",ret);
   return "no data";
  }
 usleep(1);
 for(;;)
  {
   ret =recv(socket,buffer+tl ,bufLen-tl,0);
   if(verbose)
    {
     if(tl+ret>200)
      printf("got %d byte -> (ttl=%d)\n",ret,ttl);
     else
      printf("got %d byte -> %s (ttl=%d)\n",ret,buffer,ttl);
    }   
   if(ret<=0)
    {
     if(ttl<=0)
      {
       *succ=false;
       return "";
      }
     usleep(st);
     st =st*2;
     ttl--;
    }
   else
    {
     tl += ret;
     t = buffer+tl-EOT_OFFSET-2;
     if(t[0]=='E')
      if(t[1]=='O')
       if(t[2]=='T')
	{
	 t[-1]='\0';
	 break;
	}
    }
  }
 *succ =true;
 
 if(tl-3 -EOT_OFFSET < size)
  {
   if(verbose)
    printf("\nTRANSFER too small: must=%d+%d got=%d\n",size,3 +EOT_OFFSET,tl);
   *succ= false;
  }
 
 return (buffer);
}


//
// Waits for a message from the server in reply on request or any other
// transaction. Not meant fror large data transfers
// called by send2Server, checkServer and logIntoServer
//
int wait4Message(int socket, char *buffer,int len, int timeout)
{
 char buffer2[1024],*t;
 int ret,tl,ttl=16,st=2000;
 struct pollfd ps={socket,POLLIN,0};

 buffer[0]='\0';
 if(verbose)
  printf("w4m: poll for data (timeout=%d)\n",timeout);
 if((ret=poll(&ps,(unsigned int) 1,timeout))<1)
  {
   printf("w3m: poll failed: %d %d/%d\n",ret,ps.revents,POLLIN);
   return(false);
  }
 
 ret=recv(socket,buffer,len,MSG_NOSIGNAL);
 if(verbose)
  printf("\tw4m buffer='%s'\n",buffer);
 if(ret>0)
  {
   tl=ret;
   while(buffer[tl-1] !='\0')
    {
     buffer2[0]='\0';
     ret=recv(socket,buffer2,1024,MSG_NOSIGNAL);
     if(verbose)
      printf("\tw4m  buffer2='%s'\n",buffer2);
     if(ret>0)
      {
       strcat(buffer,buffer2);
       tl +=ret;
      }
     else
      {
       if(ttl<=0)
	break;
       usleep(st);
       st =st*2;
       ttl--;
      }
    }
   if(verbose)
    printf("final buffer /%s/\n",buffer);
   return(true);
  }
 return(false);
}

//
// Interrupt driven (in opsoite to poll driven) communication functions
// Currently not used. Will be neccessary for the OSX port.
// 
#ifdef NO_POLL
#ifdef MACOSX
void setGotData(int val)
#else
void setGotData(int val, siginfo_t *info,void * ptr)
#endif
{
 char tmp[64];
 int anz;

 if(val == SIGIO)
  {
   anz =recv(mySocket,tmp,63,MSG_PEEK);
   dataAvailable =true;
  }
}
void timerStop (int value)
{
 timerStopped=true;
}
#endif

//
// signal handler for the SIGBPIE signal
//
void pipebroken (int value)
{
 if(verbose)
  printf("\n\n!!!!!!!!!!\n PIPE BROKEN %d %c\n\n\n",value,7);
}




//
// send a request (in 'cmd') to the server and wait for the answer
// Return the answer in 'buffer'
//
int send2Server(char *cmd, char *buffer, int bufLen)
{
 send2Socket(mySocket,cmd);
 
 if(!wait4Message(mySocket, buffer,bufLen,1000))
  {
   strcpy(buffer,"timeout");
   return false;
  }
 return true;
}



//
// request the answer from the server for the ticket 'ticket'
// The answer will be in buffer
//
int getAnswer(char **buffer, char * ticket)
{
 char request[256];
 int retries = 8;
 int succ;
 int slp[8] = {500000,500000,1000000,1000000,1500000,2000000,2500000,3000000};




 sprintf(request,"getAnswer %s",ticket);
 send2Socket(mySocket,request);
 *buffer = wait4Answer(mySocket,&succ);
 if(!succ)
  for(int i=0 ; i<retries; i++)
   {
    if(verbose)
     printf("SLEEP %d us until reget %d\n",slp[i],i+1);
    usleep(slp[i]);
    cleanMessagePipe(mySocket);
    sprintf(request,"getAnswer %s",ticket);
    send2Socket(mySocket,request);
    *buffer = wait4Answer(mySocket,&succ);
    if(succ)
     break;
   }
 return succ;
}




// 
// Checks if the Server is up
// return a pointer on the string 'up' or 'down'
//
char * checkServer()
{
 char *t;
 char buffer[256];
 int ret;

 if(verbose)
  {
   char txt[256];
   time_t tim=time(NULL);
   strftime(txt,255,"%Y%m%d %H %M %S",localtime(&tim));   
   if(verbose)
    printf("checking Server @ %s\n",txt);
  }

 struct sockaddr_in address[4];
 int ip ;
#ifdef MACOSX
 int len;
#else
 socklen_t len;
#endif
 // address.sa_data = buffer;
 len = 4*sizeof(struct sockaddr);
 getpeername(mySocket, (struct sockaddr *)&address,&len);
 ip = (int) address->sin_addr.s_addr;
 if(verbose)
  printf("socket is connected to %d.%d.%d.%d\n",
	 (ip & 0xFF000000)>>24 ,
	 (ip & 0xFF0000)>>16 ,
	 (ip & 0xFF00)>> 8 ,
	 (ip & 0xFF) );

 send2Socket(mySocket,"ServerUp?");
 ret = wait4Message(mySocket,buffer,255,4000);
 
 if(ret=false || strcmp(buffer,"ServerIsUp")!=0)
  {
   if(verbose)
    printf("SERVER IS UNREACHABLE (%s)\n",buffer);
   send2DamagedSocket(mySocket,"exit");
   closeSocket();
   //   socketsUp =false;
   return "down";
  }
 return "up";
}

//
// Closes the communcation sockets
//
int closeSocket ()
{
 int ret;
 if(mySocket >=0)
  {
   ret = close(mySocket);
   if(verbose)
    printf("mysocket closed ret=%d\n",ret);
  }
 mySocket=-1;
 return(0);
}

//
// closes the whole app
//
void closeApp(int value)
{
 send2Socket(mySocket,"exit");

 closeSocket();
 if(verbose)
  printf("Terminator-Client terminates %d\n",value);
 usleep(500000);
 exit(0);
}
